/*
 * 
 *  SPI-IR camera with Bluetooth output
 *  designed by Alberti's Window for
 *  Ricardo Nemirovsky, San Diego State University
 *  June-July etc. 2018
 *  
 *  Paul Antonucci
 * 
 *  
*/

#include <bluefruit.h>
#include <SPI.h>

// BLE Service
BLEDis  bledis;
BLEUart bleuart;
BLEBas  blebas;

// Software Timer for blinking RED LED
SoftwareTimer blinkTimer;
// connected status - globally accessible
int stest_connected = 0;
int camera_connected = 0;





// SPI area
// include the SPI library:
// set pin 28 as the Chip select for the SPI interface to the PixArt7025:
// allocate space for the raw SPI data


const int slaveSelectPin = 28;
const int buttonInput = 4;        // This one is the Toggle Switch next to the USB iput
const int buttonInput2 = 5;       // This one is the momentary push-button switch
const int FRAMES_PER_PACKET = 1;  // possibly up to four
const int PACKET_SIZE = 20;
byte bulk_values[255];
uint8_t transmit_buffer[25] = "123456789012345678901234";  // initialize so no undefined data at end.
const unsigned long int FRAME_INTERVAL = 100;


//int get_PixArt_data(*x_pos1, *y_pos1, *x_pos2, *y_pos2);



void setup() {
  int i, return_value, return_value2;
  // init pin on board
  pinMode (slaveSelectPin, OUTPUT);
  pinMode (buttonInput, INPUT);
  pinMode (buttonInput2, INPUT);
  digitalWrite(slaveSelectPin, HIGH);  // de-select Pixart to start - LOW is chip selected
  
  // Uart Serial communications
  Serial.begin(115200);
  Serial.println("Bluefruit52 IR Camera - Paul Antonucci for Ricardo Nemirovsky");
  Serial.println("---------------------------\n");
  
  // initialize SPI:  
  SPI.begin();
  SPI.beginTransaction(SPISettings(4000000, LSBFIRST, SPI_MODE3));  
  // setting max speed to 4Mhz gives 250ns/pulse - setting it to anywhere
  // from 6 up gives about 150ns/clock. There's no "in between"

    //task 1 - initialize Pixart Chip
    Serial.println("Initializing PIXART . . . I hope . . .  ");
    initialize_PIXART();
    
    //task 2 - set test mode  -- take this out for test
    interface_write(0xef,0x01); // select register bank 1
    interface_write(0x2b,0x00); // set test mode (second byte) on: off=0, mode 5=16 objects, 6 = 4 object, 7 = 2 moving objects 
    // probably need to activate this . . .  with
    interface_write(0xef,0x00); // # select register bank 0
    interface_write(0x25,0x05); //# no documentation on this . . .  see page 19 of tech spec
    interface_write(0x30,0x00); //# select register bank 1
    interface_write(0x30,0x01); //# set test mode on

  // NExt step - check to see if the IR Camera is available - by looking for its product ID bytes

    interface_write(0xef,0x00); // # select register bank 0
    for (i=0; i<2;i++){
        return_value = interface_read(0x02,0x00); //# product ID should be in register 2, and 3 of bank 0. 

        //Serial.print("Set of three");Serial.print(set_of_three[0],HEX);Serial.print(set_of_three[1],HEX);Serial.println(set_of_three[2],HEX);
        Serial.print("Return Value: (third byte, should = 0x25) ");Serial.println(return_value,HEX);

        return_value2 = interface_read(0x03,0x00);// Product ID should be 25, 7 in hex, 37, 112 in decimal       
        //Serial.print("Set of three for second register ");Serial.print(set_of_three[0],HEX);Serial.print(set_of_three[1],HEX);Serial.println(set_of_three[2],HEX);
        Serial.print("Return Value: (third byte, should = 0x70) ");Serial.println(return_value2,HEX);
        Serial.print("Index =  ");Serial.println(i);
        if(return_value == 0x25) camera_connected = 1;
    }        


    


  // Now the Bluetooth code
  // Initialize blinkTimer for 1000 ms and start it
  blinkTimer.begin(500, blink_timer_callback);
  blinkTimer.start();
  // Setup the BLE LED to be enabled on CONNECT
  // Note: This is actually the default behaviour, but provided
  // here in case you want to control this LED manually via PIN 19
  Bluefruit.autoConnLed(true);
  // Config the peripheral connection with maximum bandwidth 
  // more SRAM required by SoftDevice
  // Note: All config***() function must be called before begin()
  Bluefruit.configPrphBandwidth(BANDWIDTH_MAX);

  Bluefruit.begin();
  Bluefruit.setTxPower(4);   // Set max power. Accepted values are: -40, -30, -20, -16, -12, -8, -4, 0, 4
  Bluefruit.setName("Bluefruit52Oct12_010");
  Bluefruit.setConnectCallback(connect_callback);
  Bluefruit.setDisconnectCallback(disconnect_callback);

  // Configure and Start Device Information Service
  bledis.setManufacturer("Alberti's Window based on Adafruit Industries");
  bledis.setModel("Bluefruit Feather54");
  bledis.begin();

  // Configure and Start BLE Uart Service
  bleuart.begin();

  // Start BLE Battery Service
  blebas.begin();
  blebas.write(100);

  // Set up and start advertising
  startAdv();

}


void startAdv(void)
{
  // Advertising packet
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();

  // Include bleuart 128-bit uuid
  Bluefruit.Advertising.addService(bleuart);

  // Secondary Scan Response packet (optional)
  // Since there is no room for 'Name' in Advertising packet
  Bluefruit.ScanResponse.addName();
  
  /* Start Advertising
   * - Enable auto advertising if disconnected
   * - Interval:  fast mode = 20 ms, slow mode = 152.5 ms
   * - Timeout for fast mode is 30 seconds
   * - Start(timeout) with timeout = 0 will advertise forever (until connected)
   * 
   * For recommended advertising interval
   * https://developer.apple.com/library/content/qa/qa1931/_index.html   
   */
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setInterval(32, 244);    // in unit of 0.625 ms
  Bluefruit.Advertising.setFastTimeout(30);      // number of seconds in fast mode
  Bluefruit.Advertising.start(0);                // 0 = Don't stop advertising after n seconds  
}


void loop() {
  // put your main code here, to run repeatedly:
   byte button;

   int x_pos1, x_pos2, y_pos1, y_pos2;
   // moved to top . . .  to be accessed below static uint8_t transmit_buffer[25] = "12345678901234567890";  // initialize so no undefined data at end.
   uint8_t buf[64];
   int count; 
   static unsigned long current_time = millis();
   static int current_frame = 0;
   static int times_through_loop = 0;
   static char str4[ ] = "some type of text \n --  various properties - and a length of 64 or so";
   int base, scratch;



  
// Housekeeping . . . 
    /* the routine below is for testing only . . . 
    while ( digitalRead(buttonInput)) {
        Serial.print("Pausing, digital input =  ");
        button = digitalRead(buttonInput);  // if it's  high, pause so I can read the inputs
        Serial.println(button);
        delay(1000);
    } */
    // Forward data from HW Serial to BLEUART
    while (Serial.available())
    {
        // This comment is from Adafruit - "Delay to wait for enough input, since we have a limited transmission buffer"
        delay(2);
        count = Serial.readBytes(buf, sizeof(buf));
        bleuart.write( buf, count );
    }
      // Forward from BLEUART to HW Serial for now, to command loop at some point
    while ( bleuart.available() )
    {
        uint8_t ch;
        ch = (uint8_t) bleuart.read();
        Serial.write(ch);
    }

// Read Frame
    if((millis()>(current_time + FRAME_INTERVAL)) && (stest_connected ==1) ) {
        // current_time += FRAME_INTERVAL;   DOes changing this eliminate the strange 
        //  " very fast readings" that happens occasionally?
        current_time = millis();


        Serial.print("Packet time ");
        // retrieve data from PixArt chip, and sort it out
        get_PixArt_data(&x_pos1, &y_pos1, &x_pos2, &y_pos2);

        // Now put in the switch status - 
        
        // Pack the data into the output array
        // this is all moved to the get_PixARt_data routine Oct 2 2018
        /*base = current_frame * (PACKET_SIZE / FRAMES_PER_PACKET);
        
        transmit_buffer[base]     = ((x_pos1 & 0x0f00)>>8)  |  ((y_pos1 & 0x0f00)>>4) ;
        transmit_buffer[base +1]  = (x_pos1 & 0x00ff);
        transmit_buffer[base +2]  = (y_pos1 & 0x00ff);
        transmit_buffer[base +3]  = ((x_pos2 & 0x0f00)>>8)  |  ((y_pos2 & 0x0f00)>>4);
        transmit_buffer[base +4]  = (x_pos2 & 0x00ff); 
        transmit_buffer[base +5]  = (y_pos2 & 0x00ff);*/
        // Now add the two status bits
        if(digitalRead(buttonInput)) transmit_buffer[19] |= 0x10;
        else transmit_buffer[19] &=0xef;

        if(digitalRead(buttonInput2)) transmit_buffer[19] |= 0x20;
        else transmit_buffer[19] &=0xdf;
                
        // Send the packet if it has all the frames we want  . . . 

        bleuart.write(transmit_buffer, PACKET_SIZE);

      }

  // Request CPU to enter low-power mode until an event/interrupt occurs
  //waitForEvent();
}



void connect_callback(uint16_t conn_handle)
{
  char central_name[32] = { 0 };
  Bluefruit.Gap.getPeerName(conn_handle, central_name, sizeof(central_name));

  Serial.print("Connected to ");
  Serial.println(central_name);
  stest_connected = 1;
}

void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.println();
  Serial.println("Disconnected");
  stest_connected = 0;
}

/**
 * Software Timer callback is invoked via a built-in FreeRTOS thread with
 * minimal stack size. Therefore it should be as simple as possible. If
 * a periodically heavy task is needed, please use Scheduler.startLoop() to
 * create a dedicated task for it.
 * 
 * More information http://www.freertos.org/RTOS-software-timer.html
 */
void blink_timer_callback(TimerHandle_t xTimerID)
{
  (void) xTimerID;
  digitalToggle(LED_RED);
}

/**
 * RTOS Idle callback is automatically invoked by FreeRTOS
 * when there are no active threads. E.g when loop() calls delay() and
 * there is no bluetooth or hw event. This is the ideal place to handle
 * background data.
 * 
 * NOTE: FreeRTOS is configured as tickless idle mode. After this callback
 * is executed, if there is time, freeRTOS kernel will go into low power mode.
 * Therefore waitForEvent() should not be called in this callback.
 * http://www.freertos.org/low-power-tickless-rtos.html
 * 
 * WARNING: This function MUST NOT call any blocking FreeRTOS API 
 * such as delay(), xSemaphoreTake() etc ... for more information
 * http://www.freertos.org/a00016.html
 */
void rtos_idle_callback(void)
{
  // Don't call any other FreeRTOS blocking API()
  // Perform background task(s) here
}



// Subroutines for communicating with the Pixart chip
void interface_write(byte register_number, byte register_value)
{
  byte set_of_three[3];
  // First set the chip select line low
    digitalWrite(slaveSelectPin, LOW);
  // a command starts with a "command" command, of zero
  set_of_three[0] = 0x00;
  set_of_three[1] = register_number;
  set_of_three[2] = register_value;
  SPI.transfer(set_of_three,3);
  // return the select line to high
    digitalWrite(slaveSelectPin, HIGH);
}


byte interface_read(byte register_number, byte register_value) 
{
  byte set_of_three[3];
  // set the chip select low
  digitalWrite(slaveSelectPin, LOW);
  // in this case the command byte is 0x80
  set_of_three[0] = 0x80;
  set_of_three[1] = register_number;
  set_of_three[2] = register_value;
  SPI.transfer(set_of_three,3);
  // set the chip select high
  digitalWrite(slaveSelectPin, HIGH);
  return(set_of_three[2]);
  
}
byte interface_bulk_read(byte register_number, byte how_many_bytes) 
{
    interface_write(0xef,register_number); //   # either 5, 9, A or B - 255, 96, 144, or 208 byte report

    // set the chip select low
    digitalWrite(slaveSelectPin, LOW);
    // the command_byte to read many bytes values = 0x81 - that's the first byte below.  The read mode is for 140 bytes; read many bytes command
    bulk_values[0] = 0x81;
    SPI.transfer(bulk_values, 145); // 144 bytes of data plus command byte and possibly more . . 
                                  // in actual application, will not need the whole set of data -- 144 bytes ~= 1500
                                  // bit times, at 4 Mhz that would be ~=400 us . . .   still not a lot of time out of 10 ms between
    digitalWrite(slaveSelectPin, HIGH);
    return(1);    // probably should have some error code if it doesn't work - but how would I know . . . 

}

void initialize_PIXART(void)
{
    interface_write(0xef,0x00);
    interface_write(0xdc,0x00);
    interface_write(0xfb,0x04);
    interface_write(0xef,0x00);
    interface_write(0x2f,0x05);
    interface_write(0x30,0x00);
    interface_write(0x30,0x01);
    interface_write(0x1f,0x00);
    interface_write(0xef,0x01);
    interface_write(0x2d,0x00);
    interface_write(0xef,0x0c);
    interface_write(0x64,0x00);
    interface_write(0x65,0x00);
    interface_write(0x66,0x00);
    interface_write(0x67,0x00);
    interface_write(0x68,0x00);
    interface_write(0x69,0x00);
    interface_write(0x6a,0x00);
    interface_write(0x6b,0x00);
    interface_write(0x6c,0x00);
    interface_write(0x71,0x00);
    interface_write(0x72,0x00);
    interface_write(0x12,0x00);
    interface_write(0x13,0x00);
    interface_write(0xef,0x00);
    interface_write(0x01, 0x01);
}


int get_PixArt_data(int* x_pos1, int* y_pos1, int* x_pos2, int*  y_pos2) {
  int k, theDog, theCat, object_size, radius,  x_coord, y_coord, average_brightness, max_brightness; 
  static const int base = 2;   // Offset into the array due to returned SPI data being displaced by 2 bytes
  
    interface_bulk_read(0x0a, 145);
    // So what I'll want to do eventually is look at all objects, and sort them based on brightness and/or size.
    // Hm.   how will that work?  AT the end of the process, I'd have indices . . .  and then 
    // the position etc. copy wouldn't be sequential, it would be first one, then the other.
    
    // IF the object size is zero, there's no object so no need to copy it
    for(k=0;k<5;k++){   
        theDog = k*9;   // the data structure is 9 bytes long, so we index into it in steps of 9.  
                        //It's much more efficient to only calculate k*9 once, and have 9 puppies 
                        // running around than to calculate each time
        object_size = bulk_values[base + theDog +1]*256 + bulk_values[base + theDog];
        x_coord = bulk_values[base + theDog + 3] *256 +bulk_values[base + theDog + 2];
        y_coord = bulk_values[base + theDog +5] *256 + bulk_values[base + theDog + 4];
        average_brightness = bulk_values[base + theDog + 6];
        max_brightness = bulk_values[base + theDog + 7];
        radius = bulk_values[base + theDog + 8] & 0x0f;

  
        Serial.print("Object number");Serial.print(k);
        Serial.print(" size= "); Serial.print(object_size);
        Serial.print(" x = ");Serial.print(x_coord);
        Serial.print(" y = "); Serial.print(y_coord);
        Serial.print(" average brightness ");Serial.print(average_brightness);
        Serial.print(" max brightness "); Serial.println(max_brightness);
        Serial.print(" radius "); Serial.println(radius);

        // Now the object needs to be placed in the transmit array. 
        // There will be 4 bytes per object: x and y position take the first three, then 
        // the forth byte is half for the brightness, and half for the size
        // for brightness, just take the top 4 bits
        // Size is a bit messy - I'll want to take the log or the square root - 
        // The square root is available as "radius", in the low 4 bits of byte[8]
        // The first three bytes are x position and y position
        
        theCat = k*4;   
        if(theCat>4) Serial.println("The Cat has gone Wild");
        transmit_buffer[theCat]     = ((x_coord & 0x0f00)>>8)  |  ((y_coord & 0x0f00)>>4) ;
        transmit_buffer[theCat +1]  = (x_coord & 0x00ff);
        transmit_buffer[theCat +2]  = (y_coord & 0x00ff);
        transmit_buffer[theCat +3]  = (average_brightness & 0xf0) | (radius);
        
        if(k ==0) {
          *x_pos1 = x_coord;
          *y_pos1 = y_coord;
        } else {
          *x_pos2 = x_coord;
          *y_pos2 = y_coord;
        }
    } 
    return(2); 
}

